(module date
  (seconds->date-list
   today
   date-list->seconds
   date<?
   day-of-week
   days-in-month
   day->string
   day->short-string
   easter-sunday
   state-holidays
   roman-catholic-holidays
   month->string
;   moon-phase BROKEN
   muttertag
   nth-day-in
   valid-date?
   valid-partial-date?
   weekend?)

(import chicken scheme)
(use srfi-1 posix)

(define (seconds->date-list secs)
  (let ((t (seconds->utc-time secs)))
    (list (+ 1900 (vector-ref t 5))
          (add1 (vector-ref t 4))
          (vector-ref t 3))))

(define (today)
 (seconds->date-list (current-seconds)))

(define (date-list->seconds d)
  (let ((t (make-vector 10 0)))
    (set! (vector-ref t 5) (- (first d) 1900))
    (set! (vector-ref t 4) (sub1 (second d)))
    (set! (vector-ref t 3) (third d))
    (utc-time->seconds t)))

(define (date<? a b)
  (if (= (first a) (first b))
      (if (= (second a) (second b))
          (< (third a) (third b))
          (< (second a) (second b)))
      (< (first a) (first b))))


(define (leap-year? y)
  (cond ((not (zero? (modulo y 4))) #f)
        ((zero? (modulo y 400)) #t)
        ((zero? (modulo y 100)) #f)
        (else #t)))

(define days-in-month-list '(#f 31 28 31 30 31 30 31 31 30 31 30 31))
;; This is probably not correct for *all* julian calendars and their
;; reforms over the centuries but at least it has been stable for the
;; last 115 years now.

(define (days-in-month date)
  (let ((m (second date)))
    (if (and (= m 2) (leap-year? (car date)))
        29
        (list-ref days-in-month-list m))))

(define (valid-date? d)
  (and-let* ((_ (every number? d))
             (_ (= (length d) 3))
             (y (first d))
             (m (second d))
             (d (third d))
             (_ (<= 1900 y))
             (_ (<= 1 m 12))
             (max-days (days-in-month m)))
            (<= 1 d max-days)))

(define (valid-partial-date? date)
  (or (null? date)
      (and-let* ((_ (<= 1 (length date) 3))
                 (filled-partial (append date (make-list (- 3 (length date)) 1))))
                (valid-date? filled-partial))))

;; Calculates the weekday of a given date using Gauss's method
;; https://en.wikipedia.org/wiki/Determination_of_the_day_of_the_week
;; Weekday of the first day of a year A:
;;
;; R(1 + 5R(A-1,4) + 4R(A-1,100) + 6R(A-1,400),7) where R(y,m) = (modulo y m)
(define (day-of-week date)
  (let* ((A (first date))
         (m (second date))
         (d (third date))
         (first-day (modulo (+ 1
                               (* 5 (modulo (sub1 A) 4))
                               (* 4 (modulo (sub1 A) 100))
                               (* 6 (modulo (sub1 A) 400)))
                           7))
         (month-days (cond
                      ((and (= m 1) (= d 1)) -1)
                      ((> m 1)
                       (+ (if (and (> m 2) (leap-year? A)) 1 0)
                          (sub1 (apply + (take (cdr days-in-month-list) (sub1 m))))))
                      (else -1))))
    ;; The gauss method reorders the weekdays, starting at sunday (0)
    ;; We however want the week to start on monday, hence the renumbering
    (let ((dow (modulo (+ first-day d month-days) 7)))
      (if (< (sub1 dow) 0) 6 (sub1 dow)))))



;; Ben Daglish's simple method, see http://www.ben-daglish.net/moon.shtml
;; Returns the phase day (0 to 29, where 0=new moon, 15=full etc.) for
;; the selected date.
#;
(define (moon-phase date)
  (let* ((lunar-phase 29530588853)
         (new-moon (date-list->seconds '(1999 8 11)))
         (now (date-list->seconds date))
         (phase (modulo (- now new-moon) lunar-phase)))
    phase))


  (define day-names
    '(("Montag" "Mo")
      ("Dienstag" "Di")
      ("Mittwoch" "Mi")
      ("Donnerstag" "Do")
      ("Freitag" "Fr")
      ("Samstag" "Sa")
      ("Sonntag" "So")))


  (define (day->string date)
    (first (list-ref day-names (day-of-week date))))

  (define (day->short-string date)
    (second (list-ref day-names (day-of-week date))))

  (define month-names
    '(#f "Januar" "Februar" "März" "April" "Mai" "Juni" "Juli" "August" "September" "Oktober" "November" "Dezember"))

  (define (month->string date)
    (list-ref month-names (second date)))

  (define (weekend? date)
    (let ((d (day-of-week date)))
      (or (= d 5) (= d 6))))

  (define (state-holidays date)
    (let ((y (first date)))
      `(((,y 1 1) . "Neujahr")
        ((,y 5 1) . "Tag der Arbeit")
        ((,y 10 3) . "Tag d. dt. Einheit")
        ((,y 12 31) . "Silvester"))))

  ;;The following is based on the expanded Gauss easter formula
  ;; https://de.wikipedia.org/wiki/Gau%C3%9Fsche_Osterformel
  (define (easter-sunday date)
    (let* ((x (first date))
           (k (quotient x 100))
           (m (- (+ 15 (quotient (+ (* 3 k) 3) 4))
                 (quotient (+ (* 8 k) 13) 25)))
           (s (- 2 (quotient (+ (* 3 k) 3) 4) ))
           (a (modulo x 19))
           (d (modulo (+ (* 19 a) m) 30))
           (r (quotient
               (quotient (+ d a) 11)
               29))
           (og (- (+ 21 d) r))
           (sz (- 7 (modulo (+ x (quotient x 4) s) 7)))
           (oe (- 7 (modulo (- og sz) 7)))
           (os (+ og oe)))
      (if (< os 32)
          (list x 3 os)
          (list x 4 (modulo os 31)))))

  (define (days+ date days)
    (let* ((seconds/day (* 60 60 24))
           (start-date (date-list->seconds date))
           (new-date (seconds->date-list (+ start-date (* seconds/day days)))))
      new-date))

  ;; the name is a bit misleading, it should be called 'days depending
  ;; on the easter date' but that's also a dumb name
  (define (roman-catholic-holidays date)
    (let* ((y (first date))
           (es (easter-sunday date))
           (christmas-day (day-of-week `(,y 12 24)))
           (4th-advent-sunday (if (= christmas-day 6)
                                  `(,y 12 24)
                                  (days+ `(,y 12 24) (- (add1 christmas-day))))))
      `(((,y 1 6) . "Heilig. Drei Könige")
        (,(days+ es -52) . "Weiberfastnacht")
        (,(days+ es -48) . "Rosenmontag")
        (,(days+ es -47) . "Faschingsdienstag")
        (,(days+ es -46) . "Aschermittwoch")
        (,(days+ es -7) . "Palmsonntag")
        (,(days+ es -3) . "Gründonnerstag")
        (,(days+ es -2) . "Karfreitag")
        (,es . "Ostersonntag")
        (,(days+ es 1) . "Ostermontag")
        (,(days+ es 39) . "Christi Himmelfahrt")
        (,(days+ es 49) . "Pfingstsonntag")
        (,(days+ es 50) . "Pfingstmontag")
        (,(days+ es 60) . "Fronleichnam")
        (,(muttertag y) . "Muttertag")
        (,(nth-day-in 1 6 y 10) . "Erntedank")
        ((,y 8 15) . "Mariä Himmelfahrt")
        ((,y 11 1) . "Allerheiligen")
        ((,y 11 11) . "Martinstag")
        (,(days+ 4th-advent-sunday -21) . "1. Advent")
        (,(days+ 4th-advent-sunday -14) . "2. Advent")
        (,(days+ 4th-advent-sunday -7) . "3. Advent")
        (,4th-advent-sunday . "4. Advent")
        ((,y 12 24) . "Heiligabend")
        ((,y 12 25) . "1. Weihnachtsfeiertag")
        ((,y 12 26) . "2. Weihnachtsfeiertag"))))

  (define (muttertag yr)
    (nth-day-in 2 6 yr 5))

  (define (nth-day-in nth day year month)
    (let* ((m (list year month))
           (num-days (days-in-month m))
           (days (filter-map
                  (lambda (d)
                    (if (= day (day-of-week `(,@m ,d)))
                        `(,@m ,d)
                        #f))
                  (iota num-days 1))))
      (list-ref days (sub1 nth))))

  ) ;; Module
